@@ -13,29 +13,27 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import base64
 import gzip
 import logging
 import math
 import os
 import re
-import requests
+import shlex
 import shutil
-import six
 import stat
 import tempfile
 import time
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
+from oslo_serialization import base64
 from oslo_utils import excutils
 from oslo_utils import imageutils
 from oslo_utils import units
+import requests
+import six
 
 from ironic_lib.common.i18n import _
-from ironic_lib.common.i18n import _LE
-from ironic_lib.common.i18n import _LI
-from ironic_lib.common.i18n import _LW
 from ironic_lib import disk_partitioner
 from ironic_lib import exception
 from ironic_lib import utils
@@ -53,10 +51,23 @@ opts = [
     cfg.StrOpt('dd_block_size',
                default='1M',
                help='Block size to use when writing to the nodes disk.'),
+    # TODO(dtantsur): remove in Ussuri, then clean up ironic from it.
     cfg.IntOpt('iscsi_verify_attempts',
                default=3,
                help='Maximum attempts to verify an iSCSI connection is '
-                    'active, sleeping 1 second between attempts.'),
+                    'active, sleeping 1 second between attempts.',
+               deprecated_for_removal=True,
+               deprecated_reason='Split into [iscsi]verify_attempts and '
+                                 'partition_detection_attempts.'),
+    cfg.IntOpt('partition_detection_attempts',
+               min=1,
+               help='Maximum attempts to detect a newly created partition. '
+                    'Defaults to iscsi_verify_attempts, will default to 3 '
+                    'when it is removed.'),
+    cfg.IntOpt('partprobe_attempts',
+               default=10,
+               help='Maximum number of attempts to try to read the '
+                    'partition.'),
 ]
 
 CONF = cfg.CONF
@@ -65,7 +76,7 @@ CONF.register_opts(opts, group='disk_utils')
 LOG = logging.getLogger(__name__)
 
 _PARTED_PRINT_RE = re.compile(r"^(\d+):([\d\.]+)MiB:"
-                              "([\d\.]+)MiB:([\d\.]+)MiB:(\w*)::(\w*)")
+                              "([\d\.]+)MiB:([\d\.]+)MiB:(\w*):(.*):(.*);")
 
 CONFIGDRIVE_LABEL = "config-2"
 MAX_CONFIG_DRIVE_SIZE_MB = 64
@@ -73,13 +84,17 @@ MAX_CONFIG_DRIVE_SIZE_MB = 64
 # Maximum disk size supported by MBR is 2TB (2 * 1024 * 1024 MB)
 MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152
 
+# Limit the memory address space to 1 GiB when running qemu-img
+QEMU_IMG_LIMITS = processutils.ProcessLimits(address_space=1 * units.Gi)
+
 
 def list_partitions(device):
     """Get partitions information from given device.
 
     :param device: The device path.
     :returns: list of dictionaries (one per partition) with keys:
-              number, start, end, size (in MiB), filesystem, flags
+              number, start, end, size (in MiB), filesystem, partition_name,
+              flags
     """
     output = utils.execute(
         'parted', '-s', '-m', device, 'unit', 'MiB', 'print',
@@ -88,14 +103,15 @@ def list_partitions(device):
         output = output.decode("utf-8")
     lines = [line for line in output.split('\n') if line.strip()][2:]
     # Example of line: 1:1.00MiB:501MiB:500MiB:ext4::boot
-    fields = ('number', 'start', 'end', 'size', 'filesystem', 'flags')
+    fields = ('number', 'start', 'end', 'size', 'filesystem', 'partition_name',
+              'flags')
     result = []
     for line in lines:
         match = _PARTED_PRINT_RE.match(line)
         if match is None:
-            LOG.warning(_LW("Partition information from parted for device "
-                            "%(device)s does not match "
-                            "expected format: %(line)s"),
+            LOG.warning("Partition information from parted for device "
+                        "%(device)s does not match "
+                        "expected format: %(line)s",
                         dict(device=device, line=line))
             continue
         # Cast int fields to ints (some are floats and we round them down)
@@ -105,6 +121,30 @@ def list_partitions(device):
     return result
 
 
+def count_mbr_partitions(device):
+    """Count the number of primary and logical partitions on a MBR
+
+    :param device: The device path.
+    :returns: A tuple with the number of primary partitions and logical
+              partitions.
+    :raise: ValueError if the device does not have a valid MBR partition
+            table.
+    """
+    # -d do not update the kernel table
+    # -s print a summary of the partition table
+    output, err = utils.execute('partprobe', '-d', '-s', device,
+                                run_as_root=True, use_standard_locale=True)
+    if 'msdos' not in output:
+        raise ValueError('The device %s does not have a valid MBR '
+                         'partition table' % device)
+    # Sample output: /dev/vdb: msdos partitions 1 2 3 <5 6 7>
+    # The partitions with number > 4 (and inside <>) are logical partitions
+    output = output.replace('<', '').replace('>', '')
+    partitions = [int(s) for s in output.split() if s.isdigit()]
+
+    return(sum(i < 5 for i in partitions), sum(i > 4 for i in partitions))
+
+
 def get_disk_identifier(dev):
     """Get the disk identifier from the disk being exposed by the ramdisk.
 
@@ -115,6 +155,7 @@ def get_disk_identifier(dev):
     http://www.syslinux.org/wiki/index.php/Comboot/chain.c32#mbr:
 
     :param dev: Path for the already populated disk device.
+    :raises OSError: When the hexdump binary is unavailable.
     :returns: The Disk Identifier.
     """
     disk_identifier = utils.execute('hexdump', '-s', '440', '-n', '4',
@@ -127,6 +168,41 @@ def get_disk_identifier(dev):
     return disk_identifier[0]
 
 
+def get_uefi_disk_identifier(dev):
+    """Get the uuid from the disk being exposed by the ramdisk.
+
+    This uuid is appended to the pxe config which will then be set as the root
+    and load the bootx64.efi file using chainloader and boot the machine.
+    This is helpful in deployments to nodes with multiple disks.
+
+    https://wiki.gentoo.org/wiki/GRUB2/Chainloading
+
+    :param dev: Path for the already populated disk device.
+    :raises InstanceDeployFailure: Image is not UEFI bootable.
+    :returns: The UUID of the partition.
+    """
+    partition_id = None
+    try:
+        report, _ = utils.execute('fdisk', '-l', dev, run_as_root=True)
+    except processutils.ProcessExecutionError as e:
+        msg = _('Failed to find the partition on the disk %s ') % e
+        LOG.error(msg)
+        raise exception.InstanceDeployFailure(msg)
+    for line in report.splitlines():
+        if line.startswith(dev) and 'EFI System' in line:
+            vals = line.split()
+            partition_id = vals[0]
+    try:
+        lsblk_output, _ = utils.execute('lsblk', '-PbioUUID', partition_id,
+                                        run_as_root=True)
+        disk_identifier = lsblk_output.split("=")[1].strip()
+        disk_identifier = disk_identifier.strip('"')
+    except processutils.ProcessExecutionError as e:
+        raise exception.InstanceDeployFailure("Image is not UEFI bootable. "
+                                              "Error: %s " % e)
+    return disk_identifier
+
+
 def is_iscsi_device(dev, node_uuid):
     """check whether the device path belongs to an iscsi device. """
 
@@ -134,15 +210,23 @@ def is_iscsi_device(dev, node_uuid):
     return iscsi_id in dev
 
 
+def is_last_char_digit(dev):
+    """check whether device name ends with a digit"""
+    if len(dev) >= 1:
+        return dev[-1].isdigit()
+    return False
+
+
 def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
                     configdrive_mb, node_uuid, commit=True,
                     boot_option="netboot", boot_mode="bios",
-                    disk_label=None):
+                    disk_label=None, cpu_arch=""):
     """Partition the disk device.
 
     Create partitions for root, swap, ephemeral and configdrive on a
     disk device.
 
+    :param dev: Path for the device to work on.
     :param root_mb: Size of the root partition in mebibytes (MiB).
     :param swap_mb: Size of the swap partition in mebibytes (MiB). If 0,
         no partition will be created.
@@ -158,6 +242,11 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     :param disk_label: The disk label to be used when creating the
         partition table. Valid values are: "msdos", "gpt" or None; If None
         Ironic will figure it out according to the boot_mode parameter.
+    :param cpu_arch: Architecture of the node the disk device belongs to.
+        When using the default value of None, no architecture specific
+        steps will be taken. This default should be used for x86_64. When
+        set to ppc64*, architecture specific steps are taken for booting a
+        partition image locally.
     :returns: A dictionary containing the partition type as Key and partition
         path as Value for the partitions created by this method.
 
@@ -173,6 +262,8 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     # the device partitions as /dev/sda1 and not /dev/sda-part1.
     if is_iscsi_device(dev, node_uuid):
         part_template = dev + '-part%d'
+    elif is_last_char_digit(dev):
+        part_template = dev + 'p%d'
     else:
         part_template = dev + '%d'
     part_dict = {}
@@ -190,11 +281,25 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
                                     boot_flag='boot')
         part_dict['efi system partition'] = part_template % part_num
 
-    if boot_mode == "bios" and boot_option == "local" and disk_label == "gpt":
+    if (boot_mode == "bios" and boot_option == "local" and disk_label == "gpt"
+        and not cpu_arch.startswith('ppc64')):
         part_num = dp.add_partition(CONF.disk_utils.bios_boot_partition_size,
                                     boot_flag='bios_grub')
         part_dict['BIOS Boot partition'] = part_template % part_num
 
+    # NOTE(mjturek): With ppc64* nodes, partition images are expected to have
+    # a PrEP partition at the start of the disk. This is an 8 MiB partition
+    # with the boot and prep flags set. The bootloader should be installed
+    # here.
+    if (cpu_arch.startswith("ppc64") and boot_mode == "bios" and
+        boot_option == "local"):
+        LOG.debug("Add PReP boot partition (8 MB) to device: "
+                  "%(dev)s for node %(node)s",
+                  {'dev': dev, 'node': node_uuid})
+        boot_flag = 'boot' if disk_label == 'msdos' else None
+        part_num = dp.add_partition(8, part_type='primary',
+                                    boot_flag=boot_flag, extra_flags=['prep'])
+        part_dict['PReP Boot partition'] = part_template % part_num
     if ephemeral_mb:
         LOG.debug("Add ephemeral partition (%(size)d MB) to device: %(dev)s "
                   "for node %(node)s",
@@ -221,10 +326,10 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
               "for node %(node)s",
               {'dev': dev, 'size': root_mb, 'node': node_uuid})
 
-    boot_val = None
-    if (boot_mode == "bios" and boot_option == "local" and
-        disk_label == "msdos"):
-        boot_val = 'boot'
+    boot_val = 'boot' if (not cpu_arch.startswith("ppc64")
+                          and boot_mode == "bios"
+                          and boot_option == "local"
+                          and disk_label == "msdos") else None
 
     part_num = dp.add_partition(root_mb, boot_flag=boot_val)
 
@@ -233,12 +338,15 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     if commit:
         # write to the disk
         dp.commit()
+        _trigger_device_rescan(dev)
     return part_dict
 
 
 def is_block_device(dev):
     """Check whether a device is block or not."""
-    attempts = CONF.disk_utils.iscsi_verify_attempts
+    attempts = (CONF.disk_utils.iscsi_verify_attempts
+                if CONF.disk_utils.partition_detection_attempts is None
+                else CONF.disk_utils.partition_detection_attempts)
     for attempt in range(attempts):
         try:
             s = os.stat(dev)
@@ -256,9 +364,15 @@ def is_block_device(dev):
     raise exception.InstanceDeployFailure(msg)
 
 
-def dd(src, dst):
+def dd(src, dst, conv_flags=None):
     """Execute dd from src to dst."""
-    utils.dd(src, dst, 'bs=%s' % CONF.disk_utils.dd_block_size, 'oflag=direct')
+    if conv_flags:
+        extra_args = ['conv=%s' % conv_flags]
+    else:
+        extra_args = []
+
+    utils.dd(src, dst, 'bs=%s' % CONF.disk_utils.dd_block_size, 'oflag=direct',
+             *extra_args)
 
 
 def qemu_img_info(path):
@@ -267,36 +381,38 @@ def qemu_img_info(path):
         return imageutils.QemuImgInfo()
 
     out, err = utils.execute('env', 'LC_ALL=C', 'LANG=C',
-                             'qemu-img', 'info', path)
+                             'qemu-img', 'info', path,
+                             prlimit=QEMU_IMG_LIMITS)
     return imageutils.QemuImgInfo(out)
 
 
 def convert_image(source, dest, out_format, run_as_root=False):
     """Convert image to other format."""
     cmd = ('qemu-img', 'convert', '-O', out_format, source, dest)
-    utils.execute(*cmd, run_as_root=run_as_root)
+    utils.execute(*cmd, run_as_root=run_as_root, prlimit=QEMU_IMG_LIMITS)
 
 
-def populate_image(src, dst):
+def populate_image(src, dst, conv_flags=None):
     data = qemu_img_info(src)
     if data.file_format == 'raw':
-        dd(src, dst)
+        dd(src, dst, conv_flags=conv_flags)
     else:
         convert_image(src, dst, 'raw', True)
 
 
-# TODO(rameshg87): Remove this one-line method and use utils.mkfs
-# directly.
-def mkfs(fs, dev, label=None):
-    """Execute mkfs on a device."""
-    utils.mkfs(fs, dev, label)
-
-
 def block_uuid(dev):
-    """Get UUID of a block device."""
+    """Get UUID of a block device.
+
+    Try to fetch the UUID, if that fails, try to fetch the PARTUUID.
+    """
     out, _err = utils.execute('blkid', '-s', 'UUID', '-o', 'value', dev,
-                              run_as_root=True,
-                              check_exit_code=[0])
+                              run_as_root=True, check_exit_code=[0])
+    if not out:
+        LOG.debug('Falling back to partition UUID as the block device UUID '
+                  'was not found while examining %(device)s',
+                  {'device': dev})
+        out, _err = utils.execute('blkid', '-s', 'PARTUUID', '-o', 'value',
+                                  dev, run_as_root=True, check_exit_code=[0])
     return out.strip()
 
 
@@ -339,16 +455,41 @@ def destroy_disk_metadata(dev, node_uuid):
                       run_as_root=True,
                       use_standard_locale=True)
     except processutils.ProcessExecutionError as e:
-        # NOTE(zhenguo): Check if --force option is supported for wipefs,
-        # if not, we should try without it.
-        if '--force' in str(e):
-            utils.execute('wipefs', '--all', dev,
-                          run_as_root=True,
-                          use_standard_locale=True)
-        else:
-            raise e
-    LOG.info(_LI("Disk metadata on %(dev)s successfully destroyed for node "
-                 "%(node)s"), {'dev': dev, 'node': node_uuid})
+        with excutils.save_and_reraise_exception() as ctxt:
+            # NOTE(zhenguo): Check if --force option is supported for wipefs,
+            # if not, we should try without it.
+            if '--force' in str(e):
+                ctxt.reraise = False
+                utils.execute('wipefs', '--all', dev,
+                              run_as_root=True,
+                              use_standard_locale=True)
+    # NOTE(TheJulia): sgdisk attempts to load and make sense of the
+    # partition tables in advance of wiping the partition data.
+    # This means when a CRC error is found, sgdisk fails before
+    # erasing partition data.
+    # This is the same bug as
+    # https://bugs.launchpad.net/ironic-python-agent/+bug/1737556
+    dd_device = 'of=%s' % dev
+    gpt_backup = get_dev_block_size(dev) - 33
+    dd_seek = 'seek=%i' % gpt_backup
+    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
+                  run_as_root=True, use_standard_locale=True)
+    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
+                  dd_seek, run_as_root=True, use_standard_locale=True)
+    # Go ahead and let sgdisk run as well.
+    utils.execute('sgdisk', '-Z', dev, run_as_root=True,
+                  use_standard_locale=True)
+
+    try:
+        utils.wait_for_disk_to_become_available(dev)
+    except exception.IronicException as e:
+        raise exception.InstanceDeployFailure(
+            _('Destroying metadata failed on device %(device)s. '
+              'Error: %(error)s')
+            % {'device': dev, 'error': e})
+
+    LOG.info("Disk metadata on %(dev)s successfully destroyed for node "
+             "%(node)s", {'dev': dev, 'node': node_uuid})
 
 
 def _get_configdrive(configdrive, node_uuid, tempdir=None):
@@ -378,7 +519,7 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
         data = configdrive
 
     try:
-        data = six.BytesIO(base64.b64decode(data))
+        data = six.BytesIO(base64.decode_as_bytes(data))
     except TypeError:
         error_msg = (_('Config drive for node %s is not base64 encoded '
                        'or the content is malformed.') % node_uuid)
@@ -414,7 +555,7 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
 def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                  image_path, node_uuid, preserve_ephemeral=False,
                  configdrive=None, boot_option="netboot", boot_mode="bios",
-                 tempdir=None, disk_label=None):
+                 tempdir=None, disk_label=None, cpu_arch="", conv_flags=None):
     """Create partitions and copy an image to the root partition.
 
     :param dev: Path for the device to work on.
@@ -424,7 +565,8 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
         no ephemeral partition will be created.
     :param ephemeral_format: The type of file system to format the ephemeral
         partition.
-    :param image_path: Path for the instance's disk image.
+    :param image_path: Path for the instance's disk image. If ``None``,
+        the root partition is prepared but not populated.
     :param node_uuid: node's uuid. Used for logging.
     :param preserve_ephemeral: If True, no filesystem is written to the
         ephemeral block device, preserving whatever content it had (if the
@@ -437,10 +579,19 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
     :param disk_label: The disk label to be used when creating the
         partition table. Valid values are: "msdos", "gpt" or None; If None
         Ironic will figure it out according to the boot_mode parameter.
+    :param cpu_arch: Architecture of the node the disk device belongs to.
+        When using the default value of None, no architecture specific
+        steps will be taken. This default should be used for x86_64. When
+        set to ppc64*, architecture specific steps are taken for booting a
+        partition image locally.
+    :param conv_flags: Flags that need to be sent to the dd command, to control
+        the conversion of the original file when copying to the host. It can
+        contain several options separated by commas.
     :returns: a dictionary containing the following keys:
         'root uuid': UUID of root partition
         'efi system partition uuid': UUID of the uefi system partition
         (if boot mode is uefi).
+        `partitions`: mapping of partition types to their device paths.
         NOTE: If key exists but value is None, it means partition doesn't
         exist.
     """
@@ -465,9 +616,10 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                                     commit=commit,
                                     boot_option=boot_option,
                                     boot_mode=boot_mode,
-                                    disk_label=disk_label)
-        LOG.info(_LI("Successfully completed the disk device"
-                     " %(dev)s partitioning for node %(node)s"),
+                                    disk_label=disk_label,
+                                    cpu_arch=cpu_arch)
+        LOG.info("Successfully completed the disk device"
+                 " %(dev)s partitioning for node %(node)s",
                  {'dev': dev, "node": node_uuid})
 
         ephemeral_part = part_dict.get('ephemeral')
@@ -480,7 +632,7 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                 _("Root device '%s' not found") % root_part)
 
         for part in ('swap', 'ephemeral', 'configdrive',
-                     'efi system partition'):
+                     'efi system partition', 'PReP Boot partition'):
             part_device = part_dict.get(part)
             LOG.debug("Checking for %(part)s device (%(dev)s) on node "
                       "%(node)s.", {'part': part, 'dev': part_device,
@@ -494,13 +646,13 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
         # partition.  Create a fat filesystem on it.
         if boot_mode == "uefi" and boot_option == "local":
             efi_system_part = part_dict.get('efi system partition')
-            mkfs(dev=efi_system_part, fs='vfat', label='efi-part')
+            utils.mkfs(fs='vfat', path=efi_system_part, label='efi-part')
 
         if configdrive_part:
             # Copy the configdrive content to the configdrive partition
-            dd(configdrive_file, configdrive_part)
-            LOG.info(_LI("Configdrive for node %(node)s successfully copied "
-                         "onto partition %(partition)s"),
+            dd(configdrive_file, configdrive_part, conv_flags=conv_flags)
+            LOG.info("Configdrive for node %(node)s successfully copied "
+                     "onto partition %(partition)s",
                      {'node': node_uuid, 'partition': configdrive_part})
 
     finally:
@@ -509,27 +661,37 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
         if configdrive_file:
             utils.unlink_without_raise(configdrive_file)
 
-    populate_image(image_path, root_part)
-    LOG.info(_LI("Image for %(node)s successfully populated"),
-             {'node': node_uuid})
+    if image_path is not None:
+        populate_image(image_path, root_part, conv_flags=conv_flags)
+        LOG.info("Image for %(node)s successfully populated",
+                 {'node': node_uuid})
+    else:
+        LOG.debug("Root partition for %s was created, but not populated",
+                  node_uuid)
 
     if swap_part:
-        mkfs(dev=swap_part, fs='swap', label='swap1')
-        LOG.info(_LI("Swap partition %(swap)s successfully formatted "
-                     "for node %(node)s"),
+        utils.mkfs(fs='swap', path=swap_part, label='swap1')
+        LOG.info("Swap partition %(swap)s successfully formatted "
+                 "for node %(node)s",
                  {'swap': swap_part, 'node': node_uuid})
 
     if ephemeral_part and not preserve_ephemeral:
-        mkfs(dev=ephemeral_part, fs=ephemeral_format, label="ephemeral0")
-        LOG.info(_LI("Ephemeral partition %(ephemeral)s successfully "
-                     "formatted for node %(node)s"),
+        utils.mkfs(fs=ephemeral_format, path=ephemeral_part,
+                   label="ephemeral0")
+        LOG.info("Ephemeral partition %(ephemeral)s successfully "
+                 "formatted for node %(node)s",
                  {'ephemeral': ephemeral_part, 'node': node_uuid})
 
     uuids_to_return = {
         'root uuid': root_part,
-        'efi system partition uuid': part_dict.get('efi system partition')
+        'efi system partition uuid': part_dict.get('efi system partition'),
     }
 
+    if cpu_arch.startswith('ppc'):
+        uuids_to_return[
+            'PReP Boot partition uuid'
+        ] = part_dict.get('PReP Boot partition')
+
     try:
         for part, part_dev in uuids_to_return.items():
             if part_dev:
@@ -537,9 +699,9 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
 
     except processutils.ProcessExecutionError:
         with excutils.save_and_reraise_exception():
-            LOG.error(_LE("Failed to detect %s"), part)
+            LOG.error("Failed to detect %s", part)
 
-    return uuids_to_return
+    return dict(partitions=part_dict, **uuids_to_return)
 
 
 def list_opts():
@@ -557,9 +719,10 @@ def _is_disk_larger_than_max_size(device, node_uuid):
     :returns: True if total disk size exceeds 2TB. Returns False otherwise.
     """
     try:
-        disksize_bytes = utils.execute('blockdev', '--getsize64', device,
-                                       use_standard_locale=True,
-                                       run_as_root=True)
+        disksize_bytes, err = utils.execute('blockdev', '--getsize64',
+                                            device,
+                                            use_standard_locale=True,
+                                            run_as_root=True)
     except (processutils.UnknownArgumentError,
             processutils.ProcessExecutionError, OSError) as e:
         msg = (_('Failed to get size of disk %(disk)s for node %(node)s. '
@@ -568,15 +731,15 @@ def _is_disk_larger_than_max_size(device, node_uuid):
         LOG.error(msg)
         raise exception.InstanceDeployFailure(msg)
 
-    disksize_mb = int(disksize_bytes) // 1024 // 1024
+    disksize_mb = int(disksize_bytes.strip()) // 1024 // 1024
 
     return disksize_mb > MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR
 
 
-def _get_labelled_partition(device, label, node_uuid):
+def _get_labelled_partition(device_path, label, node_uuid):
     """Check and return if partition with given label exists
 
-    :param device: The device path.
+    :param device_path: The device path.
     :param label: Partition label
     :param node_uuid: UUID of the Node. Used for logging.
     :raises: InstanceDeployFailure, if any disk partitioning related
@@ -585,27 +748,42 @@ def _get_labelled_partition(device, label, node_uuid):
               returns None.
     """
     try:
-        utils.execute('partprobe', device, run_as_root=True)
-        label_arg = 'LABEL=%s' % label
-        output, err = utils.execute('blkid', '-o', 'device', device,
-                                    '-t', label_arg, check_exit_code=[0, 2],
+        utils.execute('partprobe', device_path, run_as_root=True,
+                      attempts=CONF.disk_utils.partprobe_attempts)
+
+        # lsblk command
+        output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
+                                    check_exit_code=[0, 1],
                                     use_standard_locale=True, run_as_root=True)
+
     except (processutils.UnknownArgumentError,
             processutils.ProcessExecutionError, OSError) as e:
         msg = (_('Failed to retrieve partition labels on disk %(disk)s '
                  'for node %(node)s. Error: %(error)s') %
-               {'disk': device, 'node': node_uuid, 'error': e})
+               {'disk': device_path, 'node': node_uuid, 'error': e})
         LOG.error(msg)
         raise exception.InstanceDeployFailure(msg)
 
+    found_part = None
     if output:
-        if len(output.split()) > 1:
-            raise exception.InstanceDeployFailure(
-                _('More than one config drive exists on device %(device)s '
-                  'for node %(node)s.')
-                % {'device': device, 'node': node_uuid})
+        for device in output.split('\n'):
+            dev = {key: value for key, value in (v.split('=', 1)
+                   for v in shlex.split(device))}
+            if not dev:
+                continue
+            if dev['LABEL'].upper() == label.upper():
+                if found_part:
+                    found_2 = '/dev/%(part)s' % {'part': dev['NAME'].strip()}
+                    found = [found_part, found_2]
+                    raise exception.InstanceDeployFailure(
+                        _('More than one partition with label "%(label)s" '
+                          'exists on device %(device)s for node %(node)s: '
+                          '%(found)s.') %
+                        {'label': label, 'device': device_path,
+                         'node': node_uuid, 'found': ' and '.join(found)})
+                found_part = '/dev/%(part)s' % {'part': dev['NAME'].strip()}
 
-        return output.rstrip()
+    return found_part
 
 
 def _is_disk_gpt_partitioned(device, node_uuid):
@@ -619,9 +797,9 @@ def _is_disk_gpt_partitioned(device, node_uuid):
     :returns: Boolean. Returns True if disk is GPT partitioned
     """
     try:
-        output = utils.execute('blkid', '-p', '-o', 'value', '-s', 'PTTYPE',
-                               device, use_standard_locale=True,
-                               run_as_root=True)
+        stdout, _stderr = utils.execute(
+            'blkid', '-p', '-o', 'value', '-s', 'PTTYPE', device,
+            use_standard_locale=True, run_as_root=True)
     except (processutils.UnknownArgumentError,
             processutils.ProcessExecutionError, OSError) as e:
         msg = (_('Failed to retrieve partition table type for disk %(disk)s '
@@ -630,7 +808,7 @@ def _is_disk_gpt_partitioned(device, node_uuid):
         LOG.error(msg)
         raise exception.InstanceDeployFailure(msg)
 
-    return 'gpt' in output
+    return (stdout.lower().strip() == 'gpt')
 
 
 def _fix_gpt_structs(device, node_uuid):
@@ -642,12 +820,10 @@ def _fix_gpt_structs(device, node_uuid):
         commands fail.
     """
     try:
-        output, err = utils.execute('partprobe', device,
-                                    use_standard_locale=True,
-                                    run_as_root=True)
+        output, _err = utils.execute('sgdisk', '-v', device, run_as_root=True)
 
-        search_str = "fix the GPT to use all of the space"
-        if search_str in err:
+        search_str = "it doesn't reside\nat the end of the disk"
+        if search_str in output:
             utils.execute('sgdisk', '-e', device, run_as_root=True)
     except (processutils.UnknownArgumentError,
             processutils.ProcessExecutionError, OSError) as e:
@@ -658,6 +834,67 @@ def _fix_gpt_structs(device, node_uuid):
         raise exception.InstanceDeployFailure(msg)
 
 
+def fix_gpt_partition(device, node_uuid):
+    """Fix GPT partition
+
+    Fix GPT table information when image is written to a disk which
+    has a bigger extend (e.g. 30GB image written on a 60Gb physical disk).
+
+    :param device: The device path.
+    :param node_uuid: UUID of the Node.
+    :raises: InstanceDeployFailure if exception is caught.
+    """
+    try:
+        disk_is_gpt_partitioned = _is_disk_gpt_partitioned(device, node_uuid)
+        if disk_is_gpt_partitioned:
+            _fix_gpt_structs(device, node_uuid)
+    except Exception as e:
+        msg = (_('Failed to fix GPT partition on disk %(disk)s '
+                 'for node %(node)s. Error: %(error)s') %
+               {'disk': device, 'node': node_uuid, 'error': e})
+        LOG.error(msg)
+        raise exception.InstanceDeployFailure(msg)
+
+
+def _trigger_device_rescan(device):
+    """Sync and trigger device rescan.
+
+    Disk partition performed via parted, when performed on a ramdisk
+    do not have to honor the fsync mechanism. In essence, fsync is used
+    on the file representing the block device, which falls to the kernel
+    filesystem layer to trigger a sync event. On a ramdisk using ramfs,
+    this is an explicit non-operation.
+
+    As a result of this, we need to trigger a system wide sync operation
+    which will trigger cache to flush to disk, after which partition changes
+    should be visible upon re-scan.
+
+    When ramdisks are not in use, this also helps ensure that data has
+    been safely flushed across the wire, such as on iscsi connections.
+
+    :param device: The block device containing paritions that is attempting
+                   to be updated.
+    """
+    # TODO(TheJulia): This helper method was broken out for re-use, and
+    # does not have an explicit test case for itself, and it should, but
+    # it's steps are fairly explicitly checked with mocks on execute in
+    # the partition code.
+    LOG.debug('Explicitly calling sync to force buffer/cache flush.')
+    utils.execute('sync')
+    # Make sure any additions to the partitioning are reflected in the
+    # kernel.
+    LOG.debug('Waiting until udev event queue is empty')
+    utils.execute('udevadm', 'settle')
+    try:
+        utils.execute('partprobe', device, run_as_root=True,
+                      attempts=CONF.disk_utils.partprobe_attempts)
+        # Also verify that the partitioning is correct now.
+        utils.execute('sgdisk', '-v', device, run_as_root=True)
+    except processutils.ProcessExecutionError as exc:
+        LOG.warning('Failed to verify partitioning after creating '
+                    'partition(s): %s', exc)
+
+
 def create_config_drive_partition(node_uuid, device, configdrive):
     """Create a partition for config drive
 
@@ -690,6 +927,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                   "device: %(dev)s for node %(node)s",
                   {'dev': device, 'size': confdrive_mb, 'node': node_uuid})
 
+        fix_gpt_partition(device, node_uuid)
         if config_drive_part:
             LOG.debug("Configdrive for node %(node)s exists at "
                       "%(part)s",
@@ -698,7 +936,6 @@ def create_config_drive_partition(node_uuid, device, configdrive):
             cur_parts = set(part['number'] for part in list_partitions(device))
 
             if _is_disk_gpt_partitioned(device, node_uuid):
-                _fix_gpt_structs(device, node_uuid)
                 create_option = '0:-%dMB:0' % MAX_CONFIG_DRIVE_SIZE_MB
                 utils.execute('sgdisk', '-n', create_option, device,
                               run_as_root=True)
@@ -706,49 +943,75 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                 # Check if the disk has 4 partitions. The MBR based disk
                 # cannot have more than 4 partitions.
                 # TODO(stendulker): One can use logical partitions to create
-                # a config drive if there are 4 primary partitions.
+                # a config drive if there are 3 primary partitions.
                 # https://bugs.launchpad.net/ironic/+bug/1561283
-                num_parts = len(list_partitions(device))
-                if num_parts > 3:
+                try:
+                    pp_count, lp_count = count_mbr_partitions(device)
+                except ValueError as e:
+                    raise exception.InstanceDeployFailure(
+                        _('Failed to check the number of primary partitions '
+                          'present on %(dev)s for node %(node)s. Error: '
+                          '%(error)s') % {'dev': device, 'node': node_uuid,
+                                          'error': e})
+                if pp_count > 3:
                     raise exception.InstanceDeployFailure(
                         _('Config drive cannot be created for node %(node)s. '
-                          'Disk uses MBR partitioning and already has '
-                          '%(parts)d primary partitions.')
-                        % {'node': node_uuid, 'parts': num_parts})
+                          'Disk (%(dev)s) uses MBR partitioning and already '
+                          'has %(parts)d primary partitions.')
+                        % {'node': node_uuid, 'dev': device,
+                           'parts': pp_count})
 
                 # Check if disk size exceeds 2TB msdos limit
                 startlimit = '-%dMiB' % MAX_CONFIG_DRIVE_SIZE_MB
                 endlimit = '-0'
                 if _is_disk_larger_than_max_size(device, node_uuid):
                     # Need to create a small partition at 2TB limit
-                    LOG.warning(_LW("Disk size is larger than 2TB for "
-                                    "node %(node)s. Creating config drive "
-                                    "at the end of the disk %(disk)s."),
+                    LOG.warning("Disk size is larger than 2TB for "
+                                "node %(node)s. Creating config drive "
+                                "at the end of the disk %(disk)s.",
                                 {'node': node_uuid, 'disk': device})
                     startlimit = (MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR -
                                   MAX_CONFIG_DRIVE_SIZE_MB - 1)
                     endlimit = MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR - 1
 
                 utils.execute('parted', '-a', 'optimal', '-s', '--', device,
-                              'mkpart', 'primary', 'ext2', startlimit,
+                              'mkpart', 'primary', 'fat32', startlimit,
                               endlimit, run_as_root=True)
+            # Trigger device rescan
+            _trigger_device_rescan(device)
 
             upd_parts = set(part['number'] for part in list_partitions(device))
             new_part = set(upd_parts) - set(cur_parts)
             if len(new_part) != 1:
                 raise exception.InstanceDeployFailure(
                     _('Disk partitioning failed on device %(device)s. '
-                      'Unable to retrive config drive partition information.')
+                      'Unable to retrieve config drive partition information.')
                     % {'device': device})
 
             if is_iscsi_device(device, node_uuid):
                 config_drive_part = '%s-part%s' % (device, new_part.pop())
+            elif is_last_char_digit(device):
+                config_drive_part = '%sp%s' % (device, new_part.pop())
             else:
                 config_drive_part = '%s%s' % (device, new_part.pop())
 
+            LOG.debug('Waiting until udev event queue is empty')
+            utils.execute('udevadm', 'settle')
+
+            # NOTE(vsaienko): check that devise actually exists,
+            # it is not handled by udevadm when using ISCSI, for more info see:
+            # https://bugs.launchpad.net/ironic/+bug/1673731
+            # Do not use 'udevadm settle --exit-if-exist' here
+            LOG.debug('Waiting for the config drive partition %(part)s '
+                      'on node %(node)s to be ready for writing.',
+                      {'part': config_drive_part, 'node': node_uuid})
+            utils.execute('test', '-e', config_drive_part,
+                          check_exit_code=[0], attempts=15,
+                          delay_on_retry=True)
+
         dd(confdrive_file, config_drive_part)
-        LOG.info(_LI("Configdrive for node %(node)s successfully "
-                     "copied onto partition %(part)s"),
+        LOG.info("Configdrive for node %(node)s successfully "
+                 "copied onto partition %(part)s",
                  {'node': node_uuid, 'part': config_drive_part})
 
     except (processutils.UnknownArgumentError,
